
__version__ = '0.7.1'

import logging

import click

from .asset.builder import AssetTool
from .tool.cmake import cmake_cli
from .tool.dfu import dfu_cli
from .tool.flasher import flash_cli, install_cli, launch_cli
from .tool.metadata import metadata_cli
from .tool.packer import pack_cli
from .tool.relocs import relocs_cli
from .tool.setup import setup_cli


@click.group()
@click.option('--debug', is_flag=True)
@click.option('-v', '--verbose', count=True)
def main(debug, verbose):
    log_format = '%(levelname)s: %(message)s'

    log_verbosity = min(verbose, 3)
    log_level = [logging.ERROR, logging.WARNING, logging.INFO, logging.DEBUG][log_verbosity]
    logging.basicConfig(level=log_level, format=log_format)


for n, c in AssetTool._commands.items():
    main.add_command(c)

main.add_command(cmake_cli)
main.add_command(flash_cli)
main.add_command(install_cli)
main.add_command(launch_cli)
main.add_command(metadata_cli)
main.add_command(pack_cli)
main.add_command(relocs_cli)
main.add_command(setup_cli)
main.add_command(dfu_cli)


@main.command(help='Print version and exit')
def version():
    print(__version__)



import logging

from .formatter import AssetFormatter


class AssetWriter:

    def __init__(self):
        self._assets = {}

    def add_asset(self, symbol, data):
        if symbol in self._assets:
            raise NameError(f'Symbol {symbol} has already been added.')
        self._assets[symbol] = data

    def _sorted(self, sort):
        if sort is None:
            return self._assets.items()
        elif sort == 'symbol':
            return sorted(self._assets.items())
        elif sort == 'size':
            return sorted(self._assets.items(), key=lambda i: len(i[1]))
        else:
            raise ValueError(f"Don't know how to sort by {sort}.")

    def _get_format(self, value, path, default='c_header'):
        if value is None:
            if path is None:
                logging.warning(f"No output filename, writing to stdout assuming {default}")
                return AssetFormatter.parse(default)
            else:
                fmt = AssetFormatter.guess(path)
                logging.info(f"Guessed output format {fmt} for {path}")
                return fmt
        else:
            return AssetFormatter.parse(value)

    def write(self, fmt=None, path=None, force=False, report=True, sort=None):
        fmt = self._get_format(fmt, path)
        assets = self._sorted(sort)
        fragments = [fmt.fragments(symbol, data) for symbol, data in assets]
        components = {key: [f[key] for f in fragments] for key in fragments[0]}
        outpaths = []

        for component, data in fmt.join(path, components).items():
            if path is None:
                print(data)
            else:
                outpath = path if component is None else path.with_suffix(f'.{component}')
                if outpath.exists() and not force:
                    raise FileExistsError(f'Refusing to overwrite {path} (use force)')
                else:
                    logging.info(f'Writing {outpath}')
                    if type(data) is str:
                        outpath.write_text(data, encoding='utf8')
                    else:
                        outpath.write_bytes(data)
                outpaths.append(outpath)

        if path and report:
            lines = [
                f'Formatter: {fmt.name}',
                'Files:', *(f'    {path}' for path in outpaths),
                'Assets:', *('    {}: {}'.format(symbol, len(data)) for symbol, data in assets),
                'Total size: {}'.format(sum(len(data) for symbol, data in assets)),
                '',
            ]
            path.with_name(path.stem + '_report.txt').write_text('\n'.join(lines))



import pathlib
import zlib

import construct
from construct import (Checksum, Const, CString, Flag, GreedyBytes,
                       GreedyRange, Hex, Int8ul, Int16ul, Int32ul, Padded,
                       Padding, Prefixed, RawCopy, Rebuild, Struct, len_, this)

DFU_SIGNATURE = b'DfuSe'
DFU_size = Rebuild(Int32ul, 0)


def DFU_file_length(ctx):
    '''Compute the entire file size + 4 bytes for CRC
    The total DFU file length is ostensibly the actual
    length in bytes of the resulting file.
    However DFU File Manager does not seem to agree,
    since it's output size is 16 bytes short.
    Since this is suspiciously the same as the suffix
    length in bytes, we omit that number to match
    DFU File Manager's output.
    '''
    size = 11        # DFU Header Length
    # size += 16     # DFU Suffix Length
    for target in ctx.targets:
        # Each target has a 274 byte header consisting
        # of the following fields:
        size += Const(DFU_SIGNATURE).sizeof()          # szSignature ('Target' in bytes)
        size += Int8ul.sizeof()                        # bAlternateSetting
        size += Int8ul.sizeof()                        # bTargetNamed
        size += Padding(3).sizeof()                    # Padding
        size += Padded(255, CString('utf8')).sizeof()  # szTargetName
        size += Int32ul.sizeof()                       # dwTargetSize
        size += Int32ul.sizeof()                       # dwNbElements
        size += DFU_target_size(target)

    return size


def DFU_target_size(ctx):
    '''Returns the size of the target binary data, plus the
    dwElementAddress header, and dwElementSize byte count.
    '''
    size = 0

    try:
        images = ctx.images
    except AttributeError:
        images = ctx['images']

    size += sum([DFU_image_size(image) for image in images])
    return size


def DFU_image_size(image):
    return len(image['data']) + Int32ul.sizeof() + Int32ul.sizeof()


DFU_image = Struct(
    'dwElementAddress' / Hex(Int32ul),     # Data offset address for image
    'data' / Prefixed(Int32ul, GreedyBytes)
)

DFU_target = Struct(
    'szSignature' / Const(b'Target'),      # DFU target identifier
    'bAlternateSetting' / Int8ul,          # Gives device alternate setting for which this image can be used
    'bTargetNamed' / Flag,                 # Boolean determining if the target is named
    Padding(3),                            # Mystery bytes!
    'szTargetName' / Padded(255, CString('utf8')),         # Target name
                                                           # DFU File Manager does not initialise this
                                                           # memory, so our file will not exactly match
                                                           # its output.
    'dwTargetSize' / Rebuild(Int32ul, DFU_target_size),    # Total size of target images
    'dwNbElements' / Rebuild(Int32ul, len_(this.images)),  # Count the number of target images
    'images' / GreedyRange(DFU_image)
)

DFU_body = Struct(
    'szSignature' / Const(DFU_SIGNATURE),  # DFU format identifier (changes on major revisions)
    'bVersion' / Const(1, Int8ul),         # DFU format revision   (changes on minor revisions)
    'DFUImageSize' / Rebuild(Int32ul, DFU_file_length),    # Total DFU file length in bytes
    'bTargets' / Rebuild(Int8ul, len_(this.targets)),      # Number of targets in the file

    'targets' / GreedyRange(DFU_target),

    'bcdDevice' / Int16ul,                 # Firmware version, or 0xffff if ignored
    'idProduct' / Hex(Int16ul),            # USB product ID or 0xffff to ignore
    'idVendor' / Hex(Int16ul),             # USB vendor ID or 0xffff to ignore
    'bcdDFU' / Const(0x011A, Int16ul),     # DFU specification number
    'ucDfuSignature' / Const(b'UFD'),      # 0x44, 0x46 and 0x55 ie 'DFU' but reversed
    'bLength' / Const(16, Int8ul)          # Length of the DFU suffix in bytes
)

DFU = Struct(
    'fields' / RawCopy(DFU_body),
    'dwCRC' / Checksum(Int32ul,            # CRC calculated over the whole file, except for itself
                       lambda data: 0xffffffff ^ zlib.crc32(data),
                       this.fields.data)
)


def display_dfu_info(parsed):
    print(f'''
Device: {parsed.fields.value.bcdDevice}
Target: {parsed.fields.value.idProduct:04x}:{parsed.fields.value.idVendor:04x}
Size: {parsed.fields.value.DFUImageSize:,} bytes
Targets: {parsed.fields.value.bTargets}''')
    for target in parsed.fields.value.targets:
        print(f'''
    Name: {target.szTargetName}
    Alternate Setting: {target.bAlternateSetting}
    Size: {target.dwTargetSize:,} bytes
    Images: {target.dwNbElements}''')
        for image in target.images:
            print(f'''
        Offset: {image.dwElementAddress}
        Size: {len(image.data):,} bytes
''')


def build(input_file, output_file, address, force=False, id_product=0x0000, id_vendor=0x0483):
    if not output_file.parent.is_dir():
        raise RuntimeError(f'Output directory "{output_file.parent}" does not exist!')
    elif output_file.is_file() and not force:
        raise RuntimeError(f'Existing output file "{output_file}", use --force to overwrite!')

    if not input_file.suffix == ".bin":
        raise RuntimeError(f'Input file "{input_file}", is not a .bin file?')

    output = DFU.build({'fields': {'value': {
        'targets': [{
            'bAlternateSetting': 0,
            'bTargetNamed': True,
            'szTargetName': 'ST...',
            'images': [{
                'dwElementAddress': address,
                'data': open(input_file, 'rb').read()
            }]
        }],
        'bcdDevice': 0,
        'idProduct': id_product,
        'idVendor': id_vendor
    }}})

    open(output_file, 'wb').write(output)


def read(input_file):
    try:
        return DFU.parse(open(input_file, 'rb').read())
    except construct.core.ConstructError as error:
        RuntimeError(f'Invalid dfu file {input_file} ({error})')


def dump(input_file, force=False):
    parsed = read(input_file)

    for target in parsed.fields.value.targets:
        target_id = target.bAlternateSetting
        for image in target.images:
            address = image.dwElementAddress
            data = image.data
            dest = str(input_file).replace('.dfu', '')
            filename = f"{dest}-{target_id}-{address}.bin"

            if pathlib.Path(filename).is_file() and not force:
                raise RuntimeError(f'Existing output file "{filename}", use --force to overwrite!')

            print(f"Dumping image at {address} to {filename} ({len(data)} bytes)")

            open(filename, 'wb').write(data)



import logging
import os
import pathlib
import re
import shutil
import stat
import subprocess
import textwrap

import click


# check environment before prompting
class SetupCommand(click.Command):
    def parse_args(self, ctx, args):
        logging.info("Checking for prerequisites...")

        # command/name/required version
        prereqs = [
            ('git --version', 'Git', None),
            ('cmake --version', 'CMake', [3, 9]),
            ('arm-none-eabi-gcc --version', 'GCC Arm Toolchain', [7, 3])
        ]

        # adjust path to dectct the VS Arm toolchain
        path = os.getenv('PATH')
        vs_dir = os.getenv('VSInstallDir')

        if vs_dir:
            path = ';'.join([path, vs_dir + 'Linux\\gcc_arm\\bin'])
            print(path)

        failed = False

        for command, name, version in prereqs:
            try:
                result = subprocess.run(command, stdout=subprocess.PIPE, text=True, shell=True, env={'PATH': path})

                version_str = ".".join([str(x) for x in version]) if version else 'any'
                found_version_str = re.search(r'[0-9]+\.[0-9\.]+', result.stdout).group(0)

                if version:
                    found_version_list = [int(x) for x in found_version_str.split('.')[:len(version)]]
                    if found_version_list < version:
                        logging.critical(f'Found {name} version {found_version_str}, {version_str} is required!')
                        failed = True

                logging.info(f'Found {name} version {found_version_str} (required {version_str})')
            except subprocess.CalledProcessError:
                logging.critical(f'Could not find {name}!')
                failed = True

        if failed:
            click.echo('\nCheck the documentation for info on installing.\nhttps://github.com/32blit/32blit-sdk#you-will-need')
            raise click.Abort()

        super().parse_args(ctx, args)


def install_sdk(sdk_path):
    click.echo('Installing SDK...')
    subprocess.run(['git', 'clone', 'https://github.com/32blit/32blit-sdk', str(sdk_path)], stdout=subprocess.DEVNULL, stderr=subprocess.DEVNULL)

    # checkout the latest release
    # TODO: could do something with the GitHub API and download the release?
    result = subprocess.run(['git', 'describe', '--tags', '--abbrev=0'], cwd=sdk_path, text=True, stdout=subprocess.PIPE, stderr=subprocess.DEVNULL)
    latest_tag = result.stdout.strip()
    result = subprocess.run(['git', 'checkout', latest_tag], cwd=sdk_path, stdout=subprocess.DEVNULL, stderr=subprocess.DEVNULL)


def vscode_config(project_path, sdk_path):
    (project_path / '.vscode').mkdir()

    open(project_path / '.vscode' / 'settings.json', 'w').write(textwrap.dedent(
        '''
        {
            "cmake.configureSettings": {
                "32BLIT_DIR": "{sdk_path}"
            },
            "C_Cpp.default.configurationProvider": "ms-vscode.cmake-tools"
        }
        '''.replace('{sdk_path}', str(sdk_path).replace('\\', '\\\\'))))

    open(project_path / '.vscode' / 'cmake-kits.json', 'w').write(textwrap.dedent(
        '''
        [
            {
                "name": "32blit",
                "toolchainFile": "{sdk_path}/32blit.toolchain"
            }
        ]
        '''.replace('{sdk_path}', str(sdk_path).replace('\\', '\\\\'))))


def visualstudio_config(project_path, sdk_path):
    open(project_path / 'CMakeSettings.json', 'w').write(textwrap.dedent(
        '''
        {
            "configurations": [
                {
                    "name": "x64-Debug",
                    "generator": "Ninja",
                    "configurationType": "Debug",
                    "inheritEnvironments": [ "msvc_x64_x64" ],
                    "buildRoot": "${projectDir}\\\\out\\\\build\\\\${name}",
                    "installRoot": "${projectDir}\\\\out\\\\install\\\\${name}",
                    "cmakeCommandArgs": "",
                    "buildCommandArgs": "",
                    "ctestCommandArgs": "",
                    "variables": [
                        {
                            "name": "32BLIT_DIR",
                            "value": "{sdk_path}",
                            "type": "PATH"
                        }
                    ]
                },
                {
                    "name": "x64-Release",
                    "generator": "Ninja",
                    "configurationType": "Release",
                    "buildRoot": "${projectDir}\\\\out\\\\build\\\\${name}",
                    "installRoot": "${projectDir}\\\\out\\\\install\\\\${name}",
                    "cmakeCommandArgs": "",
                    "buildCommandArgs": "",
                    "ctestCommandArgs": "",
                    "inheritEnvironments": [ "msvc_x64_x64" ],
                    "variables": [
                        {
                            "name": "32BLIT_DIR",
                            "value": "{sdk_path}",
                            "type": "PATH"
                        }
                    ]
                },
                {
                    "name": "32Blit-Debug",
                    "generator": "Ninja",
                    "configurationType": "Debug",
                    "buildRoot": "${projectDir}\\\\out\\\\build\\\\${name}",
                    "installRoot": "${projectDir}\\\\out\\\\install\\\\${name}",
                    "cmakeCommandArgs": "",
                    "buildCommandArgs": "",
                    "ctestCommandArgs": "",
                    "inheritEnvironments": [ "gcc-arm" ],
                    "variables": [],
                    "cmakeToolchain": "{sdk_path}\\\\32blit.toolchain",
                    "intelliSenseMode": "linux-gcc-arm"
                },
                {
                    "name": "32Blit-Release",
                    "generator": "Ninja",
                    "configurationType": "Release",
                    "buildRoot": "${projectDir}\\\\out\\\\build\\\\${name}",
                    "installRoot": "${projectDir}\\\\out\\\\install\\\\${name}",
                    "cmakeCommandArgs": "",
                    "buildCommandArgs": "",
                    "ctestCommandArgs": "",
                    "cmakeToolchain": "{sdk_path}\\\\32blit.toolchain",
                    "inheritEnvironments": [ "gcc-arm" ],
                    "variables": [],
                    "intelliSenseMode": "linux-gcc-arm"
                }
            ]
        }
        '''.replace('{sdk_path}', str(sdk_path).replace('\\', '\\\\'))))


@click.command('setup', help='Setup a project', cls=SetupCommand)
@click.option('--project-name', prompt=True)
@click.option('--author-name', prompt=True)
@click.option('--sdk-path', type=pathlib.Path, default=lambda: os.path.expanduser('~/32blit-sdk'), prompt='32Blit SDK path')
@click.option('--git/--no-git', prompt='Initialise a Git repository?', default=True)
@click.option('--vscode/--no-vscode', prompt='Create VS Code configuration?', default=True)
@click.option('--visualstudio/--no-visualstudio', prompt='Create Visual Studio configuration?')
def setup_cli(project_name, author_name, sdk_path, git, vscode, visualstudio):
    if not (sdk_path / '32blit.toolchain').exists():
        click.confirm(f'32Blit SDK not found at "{sdk_path}", would you like to install it?', abort=True)
        install_sdk(sdk_path)

    project_name_clean = re.sub(r'[^a-z0-9]+', '-', project_name.lower()).strip('-')
    project_path = pathlib.Path.cwd() / project_name_clean

    if project_path.exists():
        logging.critical(f'A project already exists at {project_path}!')
        raise click.Abort()

    # get the boilerplate
    click.echo('Downloading boilerplate...')
    subprocess.run(['git', 'clone', '--depth', '1', 'https://github.com/32blit/32blit-boilerplate', str(project_path)], stdout=subprocess.DEVNULL, stderr=subprocess.DEVNULL)

    # de-git it (using the template on GitHub also removes the history)
    def remove_readonly(func, path, _):
        os.chmod(path, stat.S_IWRITE)
        func(path)

    shutil.rmtree(pathlib.Path(project_name_clean) / '.git', onerror=remove_readonly)

    # do some editing
    cmakelists = open(project_path / 'CMakeLists.txt').read()
    cmakelists = cmakelists.replace('project(game)', f'project({project_name_clean})')
    open(project_path / 'CMakeLists.txt', 'w').write(cmakelists)

    metadata = open(project_path / 'metadata.yml').read()
    metadata = metadata.replace('game title', project_name).replace('you', author_name)
    open(project_path / 'metadata.yml', 'w').write(metadata)

    licence = open(project_path / 'LICENSE').read()
    licence = licence.replace('<insert your name>', author_name)
    open(project_path / 'LICENSE', 'w').write(licence)

    # re-git it if we want a git repo
    if git:
        subprocess.run(['git', 'init'], cwd=project_path, stdout=subprocess.DEVNULL, stderr=subprocess.DEVNULL)
        subprocess.run(['git', 'add', '.'], cwd=project_path, stdout=subprocess.DEVNULL, stderr=subprocess.DEVNULL)
        subprocess.run(['git', 'commit', '-m', 'Initial commit'], cwd=project_path, stdout=subprocess.DEVNULL, stderr=subprocess.DEVNULL)

    if vscode:
        vscode_config(project_path, sdk_path)
    if visualstudio:
        visualstudio_config(project_path, sdk_path)

    click.echo(f'\nYour new project has been created in: {project_path}!')

    click.echo(f'If using CMake directly, make sure to pass -DCMAKE_TOOLCHAIN_FILE={sdk_path / "32blit.toolchain"} (building for the device)\nor -D32BLIT_DIR={sdk_path} when calling cmake.')



import logging
import pathlib
import textwrap

import click

from ..asset.formatter import AssetFormatter
from ..core.yamlloader import YamlLoader


class CMake(YamlLoader):

    def run(self, config, cmake, output):
        self.setup_for_config(config, output)

        if 'title' in self.config and 'description' in self.config:
            logging.info('Detected metadata config')
            self.run_for_metadata_config(cmake)
        else:
            logging.info('Detected asset config')
            self.run_for_asset_config(cmake)

    def run_for_metadata_config(self, cmake):
        all_inputs = []

        if 'splash' in self.config:
            file = pathlib.Path(self.config['splash']['file'])
            if file.is_absolute():
                all_inputs += [file]
            else:
                all_inputs += list(self.working_path.glob(str(file)))

        if 'icon' in self.config:
            file = pathlib.Path(self.config['icon']['file'])
            if file.is_absolute():
                all_inputs += [file]
            else:
                all_inputs += list(self.working_path.glob(str(file)))

        all_inputs = '\n    '.join(f'"{x}"'.replace('\\', '/') for x in all_inputs)
        title = self.config['title'].replace('"', r'\"')
        author = self.config['author'].replace('"', r'\"')
        description = self.config['description'].replace('"', r'\"')

        url = self.config.get('url', '')
        category = self.config.get('category', 'none')

        result = textwrap.dedent(
            '''\
            # Auto Generated File - DO NOT EDIT!
            set(METADATA_DEPENDS
                {inputs}
            )
            set(METADATA_TITLE "{title}")
            set(METADATA_AUTHOR "{author}")
            set(METADATA_DESCRIPTION "{description}")
            set(METADATA_VERSION "{config[version]}")
            set(METADATA_URL "{url}")
            set(METADATA_CATEGORY "{category}")
            '''
        ).format(
            inputs=all_inputs,
            config=self.config,
            title=title,
            author=author,
            description=description,
            url=url,
            category=category,
        )

        cmake.write_text(result)

    def run_for_asset_config(self, cmake):
        all_inputs = []
        all_outputs = []

        # Top level of our config is filegroups and general settings
        for target, options in self.config.items():
            target = pathlib.Path(target)

            try:
                output_formatter = AssetFormatter.guess(target)
            except TypeError:
                logging.warning(f'Unable to guess type of {target}, assuming raw/binary')
                output_formatter = AssetFormatter.parse('raw_binary')

            for suffix in output_formatter.components:
                if suffix is None:
                    all_outputs.append(self.destination_path / target.name)
                else:
                    all_outputs.append(self.destination_path / target.with_suffix(f'.{suffix}').name)

            # Strip high-level options from the dict
            # Leaving just file source globs
            for key in ('prefix', 'type'):
                options.pop(key, None)

            for file_glob, file_options in options.items():
                # Rewrite a single string option to `name: option`
                # This handles: `inputfile.bin: filename` entries
                if type(file_options) is str:
                    file_options = {'name': file_options}

                # Handle both an array of options dicts or a single dict
                if type(file_options) is not list:
                    file_options = [file_options]

                input_files = []
                # Parse the options for any references to input files
                for file_opts in file_options:
                    for key, value in file_opts.items():
                        if key in ('palette') and type(value) is str:
                            input_files += list(self.working_path.glob(value))

                # Treat the input string as a glob, and get an input filelist
                if type(file_glob) is str:
                    input_files += list(self.working_path.glob(file_glob))
                else:
                    input_files += [file_glob]
                if len(input_files) == 0:
                    logging.warning(f'Input file(s) not found {self.working_path / file_glob}')
                    continue

                all_inputs += input_files

        all_inputs = '\n    '.join(f'"{x}"'.replace('\\', '/') for x in all_inputs)
        all_outputs = '\n    '.join(f'"{x}"'.replace('\\', '/') for x in all_outputs)

        result = textwrap.dedent(
            '''\
            # Auto Generated File - DO NOT EDIT!
            set(ASSET_DEPENDS
                {inputs}
            )

            set(ASSET_OUTPUTS
                {outputs}
            )
            '''
        ).format(
            inputs=all_inputs,
            outputs=all_outputs,
        )

        cmake.write_text(result)


@click.command('cmake', help='Generate CMake configuration for the asset packer')
@click.option('--config', type=pathlib.Path, help='Asset config file')
@click.option('--cmake', type=pathlib.Path, help='Output CMake file', required=True)
@click.option('--output', type=pathlib.Path, help='Name for output file(s) or root path when using --config')
def cmake_cli(config, cmake, output):
    CMake().run(config, cmake, output)



from ..builder import AssetBuilder, AssetTool

binary_typemap = {
    'binary': {
        '.bin': True,
        '.raw': True,
    },
    'csv': {
        '.csv': True
    }
}


def csv_to_list(input_data, base):
    if type(input_data) == bytes:
        input_data = input_data.decode('utf-8')

    # Strip leading/trailing whitespace
    input_data = input_data.strip()

    # Replace '1, 2, 3' to '1,2,3', might as well do it here
    input_data = input_data.replace(' ', '')

    # Split out into rows on linebreak
    input_data = input_data.split('\n')

    # Split every row into columns on the comma
    input_data = [row.split(',') for row in input_data]

    # Flatten our rows/cols 2d array into a 1d array of bytes
    # Might as well do the int conversion here, to save another loop
    return [int(col, base) for row in input_data for col in row if col != '']


@AssetBuilder(typemap=binary_typemap)
def raw(data, subtype):
    if subtype == 'csv':
        return bytes(csv_to_list(data, base=10))
    else:
        return data


@AssetTool(raw, 'Convert raw/binary or csv data for 32Blit')
def raw_cli(input_file, input_type):
    return raw.from_file(input_file, input_type)



import textwrap

from ..formatter import AssetFormatter

wrapper = textwrap.TextWrapper(
    initial_indent='    ', subsequent_indent='    ', width=80
)


def c_initializer(data):
    if type(data) is str:
        data = data.encode('utf-8')
    values = ', '.join(f'0x{c:02x}' for c in data)
    return f' = {{\n{wrapper.fill(values)}\n}}'


def c_declaration(types, symbol, data=None):
    return textwrap.dedent(
        '''\
        {types} uint8_t {symbol}[]{initializer};
        {types} uint32_t {symbol}_length{size};
        '''
    ).format(
        types=types,
        symbol=symbol,
        initializer=c_initializer(data) if data else '',
        size=f' = sizeof({symbol})' if data else '',
    )


def c_boilerplate(data, include, header=True):
    lines = ['// Auto Generated File - DO NOT EDIT!']
    if header:
        lines.append('#pragma once')
    lines.append(f'#include <{include}>')
    lines.append('')
    lines.extend(data)
    return '\n'.join(lines)


@AssetFormatter(extensions=('.hpp', '.h'))
def c_header(symbol, data):
    return {None: c_declaration('inline const', symbol, data)}


@c_header.joiner
def c_header(path, fragments):
    return {None: c_boilerplate(fragments[None], include="cstdint", header=True)}


@AssetFormatter(components=('hpp', 'cpp'), extensions=('.cpp', '.c'))
def c_source(symbol, data):
    return {
        'hpp': c_declaration('extern const', symbol),
        'cpp': c_declaration('const', symbol, data),
    }


@c_source.joiner
def c_source(path, fragments):
    include = path.with_suffix('.hpp').name
    return {
        'hpp': c_boilerplate(fragments['hpp'], include='cstdint', header=True),
        'cpp': c_boilerplate(fragments['cpp'], include=include, header=False),
    }



import functools
import logging
import pathlib
import textwrap

import click

from ..core.blitserial import BlitSerial

block_size = 64 * 1024


@click.group('flash', help='Flash a binary or save games/files to 32Blit')
@click.option('ports', '--port', multiple=True, help='Serial port')
@click.pass_context
def flash_cli(ctx, ports):
    ctx.obj = ports


def serial_command(fn):
    """Set up and tear down serial connections."""

    @click.option('ports', '--port', multiple=True, help='Serial port')
    @click.pass_context
    @functools.wraps(fn)
    def _decorated(ctx, ports, **kwargs):
        if ctx.obj:
            ports = ctx.obj + ports
        if not ports or ports[0].lower() == 'auto':
            ports = BlitSerial.find_comport()

        for port in ports:
            with BlitSerial(port) as sp:
                fn(sp, **kwargs)

    return _decorated


@flash_cli.command(help="Deprecated: use '32blit install' instead. Copy a file to SD card")
@serial_command
@click.option('--file', type=pathlib.Path, required=True, help='File to save')
@click.option('--directory', type=str, default='/', help='Target directory')
def save(blitserial, file, directory):
    blitserial.send_file(file, 'sd', directory=directory)


@flash_cli.command(help="Deprecated: use '32blit install' instead. Install a .blit to flash")
@serial_command
@click.option('--file', type=pathlib.Path, required=True, help='File to save')
def flash(blitserial, file):
    blitserial.send_file(file, 'flash')


# TODO: options should be mutually exclusive
@flash_cli.command(help="Delete a .blit from flash by block index or offset")
@serial_command
@click.option('--offset', type=int, help='Flash offset of game to delete')
@click.option('--block', type=int, help='Flash block of game to delete')
def delete(blitserial, offset, block):
    if offset is None:
        offset = block * block_size
    blitserial.erase(offset)


@flash_cli.command(help="List .blits installed in flash memory")
@serial_command
def list(blitserial):
    for meta, offset, size in blitserial.list():
        offset_blocks = offset // block_size
        size_blocks = (size - 1) // block_size + 1

        print(f"Game at block {offset_blocks}")
        print(f"Size: {size_blocks:3d} blocks ({size / 1024:.1f}kB)")

        if meta is not None:
            print(textwrap.dedent(f"""\
                Title:       {meta.data.title}
                Description: {meta.data.description}
                Version:     {meta.data.version}
                Author:      {meta.data.author}
            """))


@flash_cli.command(help="Not implemented")
@serial_command
def debug(blitserial):
    pass


@flash_cli.command(help="Reset 32Blit")
@serial_command
def reset(blitserial):
    logging.info('Resetting your 32Blit...')
    blitserial.reset()


@flash_cli.command(help="Get current runtime status of 32Blit")
@serial_command
def info(blitserial):
    logging.info('Getting 32Blit run status...')
    print(f'Running: {blitserial.status}')


@click.command('install', help="Install files to 32Blit")
@serial_command
@click.argument("source", type=pathlib.Path, required=True)
@click.argument("destination", type=pathlib.PurePosixPath, default=None, required=False)
@click.option("--launch/--no-launch", default=False)
def install_cli(blitserial, source, destination, launch):
    if destination is None and source.suffix.lower() == '.blit':
        drive = 'flash'
    else:
        drive = 'sd'
        if destination is None:
            destination = pathlib.PurePosixPath('/')
        elif not destination.is_absolute():
            destination = pathlib.PurePosixPath('/') / destination

    blitserial.send_file(source, drive, destination, launch)


@click.command('launch', help="Launch a game/file on 32Blit")
@serial_command
@click.argument("filename", type=str, required=True)
def launch_cli(blitserial, filename):
    blitserial.launch(filename)



import io
import logging
import pathlib

import click
from PIL import Image

from ...core.palette import Colour, Palette
from ...core.struct import struct_blit_image
from ..builder import AssetBuilder, AssetTool

image_typemap = {
    'image': {
        '.png': True,
        '.gif': True,
    }
}


@AssetBuilder(typemap=image_typemap)
def image(data, subtype, palette=None, transparent=None, strict=False, packed=True):
    if palette is None:
        palette = Palette()
    else:
        palette = Palette(palette)
    if transparent is not None:
        transparent = Colour(transparent)
        p = palette.set_transparent_colour(*transparent)
        if p is not None:
            logging.info(f'Found transparent {transparent} in palette')
        else:
            logging.warning(f'Could not find transparent {transparent} in palette')
    # Since we already have bytes, we need to pass PIL an io.BytesIO object
    image = Image.open(io.BytesIO(data)).convert('RGBA')
    image = palette.quantize_image(image, transparent=transparent, strict=strict)
    return struct_blit_image.build({
        'type': None if packed else 'RW',  # None means let the compressor decide
        'data': {
            'width': image.size[0],
            'height': image.size[1],
            'palette': palette.tostruct(),
            'pixels': image.tobytes(),
        },
    })


@AssetTool(image, 'Convert images/sprites for 32Blit')
@click.option('--palette', type=pathlib.Path, help='Image or palette file of colours to use')
@click.option('--transparent', type=Colour, default=None, help='Transparent colour')
@click.option('--packed', type=click.Choice(['yes', 'no'], case_sensitive=False), default='yes', help='Pack into bits depending on palette colour count')
@click.option('--strict/--no-strict', default=False, help='Reject colours not in the palette')
def image_cli(input_file, input_type, packed, **kwargs):
    packed = (packed.lower() == 'yes')
    return image.from_file(input_file, input_type, packed=packed, **kwargs)



import io
import logging
import pathlib
from datetime import datetime

import click
from construct.core import StreamError
from PIL import Image

from ..asset.builder import AssetBuilder
from ..core.struct import (blit_game, blit_game_with_meta,
                           blit_game_with_meta_and_relo, blit_icns,
                           struct_blit_image, struct_blit_pixel)
from ..core.yamlloader import YamlLoader


class Metadata(YamlLoader):

    def prepare_image_asset(self, name, config, working_path):
        image_file = pathlib.Path(config.get('file', ''))
        if not image_file.is_file():
            image_file = working_path / image_file
        if not image_file.is_file():
            raise ValueError(f'{name} "{image_file}" does not exist!')
        return AssetBuilder._by_name['image'].from_file(image_file, None)

    def blit_image_to_pil(self, image):
        data = b''.join(struct_blit_pixel.build(image.data.palette[i]) for i in image.data.pixels)
        return Image.frombytes("RGBA", (image.data.width, image.data.height), data)

    def build_icns(self, config, working_path):
        image_file = pathlib.Path(config.get('file', ''))
        if not image_file.is_file():
            image_file = working_path / image_file
        if not image_file.is_file():
            raise ValueError(f'splash "{image_file}" does not exist!')

        blit_icon = Image.frombytes("P", (22, 22), b'\x01\x01\x01\x01\x01\x01\x01\x01\x01\x01\x00\x00\x01\x01\x01\x01\x01\x01\x01\x01\x02\x02\x01\x01\x01\x01\x01\x01\x01\x01\x01\x01\x00\x00\x01\x01\x01\x01\x01\x01\x01\x01\x02\x02\x00\x00\x00\x00\x00\x00\x00\x00\x01\x01\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x02\x02\x00\x00\x00\x00\x00\x00\x00\x00\x01\x01\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x02\x02\x01\x01\x01\x01\x01\x01\x01\x01\x01\x01\x00\x00\x01\x01\x01\x01\x01\x01\x01\x01\x02\x02\x01\x01\x01\x01\x01\x01\x01\x01\x01\x01\x00\x00\x01\x01\x01\x01\x01\x01\x01\x01\x02\x02\x00\x00\x00\x00\x00\x00\x00\x00\x01\x01\x00\x00\x01\x01\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x01\x01\x00\x00\x01\x01\x00\x00\x00\x00\x00\x00\x00\x00\x01\x01\x01\x01\x01\x01\x01\x01\x01\x01\x00\x00\x01\x01\x01\x01\x01\x01\x01\x01\x01\x01\x01\x01\x01\x01\x01\x01\x01\x01\x01\x01\x00\x00\x01\x01\x01\x01\x01\x01\x01\x01\x01\x01\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x01\x01\x00\x00\x00\x00\x00\x00\x00\x00\x01\x01\x00\x00\x02\x02\x00\x00\x01\x01\x00\x00\x01\x01\x00\x00\x00\x00\x00\x00\x00\x00\x01\x01\x00\x00\x02\x02\x00\x00\x01\x01\x00\x00\x01\x01\x01\x01\x01\x01\x01\x01\x00\x00\x01\x01\x00\x00\x00\x00\x00\x00\x01\x01\x01\x01\x01\x01\x01\x01\x01\x01\x01\x01\x00\x00\x01\x01\x00\x00\x00\x00\x00\x00\x01\x01\x01\x01\x02\x02\x00\x00\x00\x00\x01\x01\x00\x00\x01\x01\x00\x00\x01\x01\x00\x00\x01\x01\x00\x00\x02\x02\x00\x00\x00\x00\x01\x01\x00\x00\x01\x01\x00\x00\x01\x01\x00\x00\x01\x01\x00\x00\x02\x02\x00\x00\x00\x00\x01\x01\x00\x00\x01\x01\x00\x00\x01\x01\x00\x00\x01\x01\x00\x00\x02\x02\x00\x00\x00\x00\x01\x01\x00\x00\x01\x01\x00\x00\x01\x01\x00\x00\x01\x01\x00\x00\x02\x02\x01\x01\x01\x01\x01\x01\x00\x00\x01\x01\x00\x00\x01\x01\x00\x00\x01\x01\x01\x01\x02\x02\x01\x01\x01\x01\x01\x01\x00\x00\x01\x01\x00\x00\x01\x01\x00\x00\x01\x01\x01\x01')
        blit_icon.putpalette([0, 0, 0, 255, 255, 255, 0, 255, 0])
        blit_icon = blit_icon.convert("RGBA")

        source = Image.open(image_file).convert("RGBA")

        image = Image.new("RGB", (128, 128), (0, 0, 0))

        image.paste(source, (0, 0))
        image.paste(blit_icon, (101, 101), blit_icon)

        image_bytes = io.BytesIO()
        image.convert("RGB").save(image_bytes, format="PNG")
        image_bytes.seek(0)

        del image

        return blit_icns.build({'data': image_bytes.read()})

    def run(self, config, icns, file, force, dump_images):
        if file is None and config is None:
            raise click.UsageError('the following arguments are required: --config and/or --file')

        if file and not file.is_file():
            raise ValueError(f'Unable to find bin file at {file}')

        icon = b''
        splash = b''

        game = None

        if file:
            bin = open(file, 'rb').read()
            try:
                game = blit_game.parse(bin)
            except StreamError:
                raise ValueError(f'Invalid 32blit binary file {file}')

        # No config supplied, so dump the game info
        if config is None:
            print(f'\nParsed:      {file.name} ({game.bin.length:,} bytes)')
            if game.relo is not None:
                print(f'Relocations: Yes ({len(game.relo.relocs)})')
            else:
                print('Relocations: No')
            if game.meta is not None:
                print('Metadata:    Yes')
                for field in ['title', 'description', 'version', 'author', 'category', 'url']:
                    print(f'{field.title()+":":13s}{getattr(game.meta.data, field)}')
                if len(game.meta.data.filetypes) > 0:
                    print('    Filetypes:   ')
                    for filetype in game.meta.data.filetypes:
                        print('       ', filetype)
                if game.meta.data.icon is not None:
                    game_icon = game.meta.data.icon
                    print(f'    Icon:        {game_icon.data.width}x{game_icon.data.height} ({len(game_icon.data.palette)} colours) ({game_icon.type})')
                    if dump_images:
                        image_icon = self.blit_image_to_pil(game_icon)
                        image_icon_filename = file.with_suffix(".icon.png")
                        image_icon.save(image_icon_filename)
                        print(f'    Dumped to:   {image_icon_filename}')
                if game.meta.data.splash is not None:
                    game_splash = game.meta.data.splash
                    print(f'    Splash:      {game_splash.data.width}x{game_splash.data.height} ({len(game_splash.data.palette)} colours) ({game_splash.type})')
                    if dump_images:
                        image_splash = self.blit_image_to_pil(game_splash)
                        image_splash_filename = file.with_suffix('.splash.png')
                        image_splash.save(image_splash_filename)
                        print(f'    Dumped to:   {image_splash_filename}')
            else:
                print('Metadata:    No')
            print('')
            return

        self.setup_for_config(config, None)

        if 'icon' in self.config:
            icon = self.prepare_image_asset('icon', self.config['icon'], self.working_path)
        else:
            raise ValueError('An 8x8 pixel icon is required!"')

        if 'splash' in self.config:
            splash = self.prepare_image_asset('splash', self.config['splash'], self.working_path)
            if icns is not None:
                if not icns.is_file() or force:
                    open(icns, 'wb').write(self.build_icns(self.config['splash'], self.working_path))
                    logging.info(f'Saved macOS icon to {icns}')
        else:
            raise ValueError('A 128x96 pixel splash is required!"')

        if not game:
            return

        title = self.config.get('title', None)
        description = self.config.get('description', '')
        version = self.config.get('version', None)
        author = self.config.get('author', None)
        url = self.config.get('url', '')
        category = self.config.get('category', 'none')
        filetypes = self.config.get('filetypes', [])
        if type(filetypes) is str:
            filetypes = filetypes.split(' ')

        if title is None:
            raise ValueError('Title is required!')

        if version is None:
            raise ValueError('Version is required!')

        if author is None:
            raise ValueError('Author is required!')

        if len(title) > 24:
            raise ValueError('Title should be a maximum of 24 characters!"')

        if len(description) > 128:
            raise ValueError('Description should be a maximum of 128 characters!')

        if len(version) > 16:
            raise ValueError('Version should be a maximum of 16 characters! eg: "v1.0.2"')

        if len(author) > 16:
            raise ValueError('Author should be a maximum of 16 characters!')

        if len(category) > 16:
            raise ValueError('Category should be a maximum of 16 characters!')

        if len(url) > 128:
            raise ValueError('URL should be a maximum of 128 characters!')

        if len(filetypes) > 0:
            for filetype in filetypes:
                if len(filetype) > 4:
                    raise ValueError('Filetype should be a maximum of 4 characters! (Hint, don\'t include the .)')

        if game.meta is not None:
            if not force:
                logging.critical(f'Refusing to overwrite metadata in {file}')
                return 1

        game.meta = {
            'data': {
                'date': datetime.now().strftime("%Y%m%dT%H%M%S"),
                'title': title,
                'description': description,
                'version': version,
                'author': author,
                'category': category,
                'filetypes': filetypes,
                'url': url,
                'icon': struct_blit_image.parse(icon),
                'splash': struct_blit_image.parse(splash)
            }
        }

        # Force through a non-optional builder if relo symbols exist
        # since we might have modified them and want to hit parser errors
        # if something has messed up
        if game.relo is not None:
            bin = blit_game_with_meta_and_relo.build(game)
        else:
            bin = blit_game_with_meta.build(game)

        logging.info(f'Adding metadata to {file}')
        open(file, 'wb').write(bin)

        return 0


@click.command('metadata', help='Tag a 32Blit .blit file with metadata')
@click.option('--config', type=pathlib.Path, help='Metadata config file')
@click.option('--icns', type=pathlib.Path, help='macOS icon output file')
@click.option('--file', type=pathlib.Path, help='Input file')
@click.option('--force', is_flag=True, help='Force file overwrite')
@click.option('--dump-images', is_flag=True, help='Dump images from metadata')
def metadata_cli(config, icns, file, force, dump_images):
    Metadata().run(config, icns, file, force, dump_images)



import pathlib

import click

from ..core import dfu


@click.group('dfu', help='Pack or unpack DFU files')
def dfu_cli():
    pass


@dfu_cli.command(help='Pack a .bin file into a DFU file')
@click.option('--input-file', type=pathlib.Path, help='Input .bin', required=True)
@click.option('--output-file', type=pathlib.Path, help='Output .dfu', required=True)
@click.option('--address', type=int, help='Offset address', default=0x08000000)
@click.option('--force/--no-force', help='Force file overwrite', default=True)
def build(input_file, output_file, address, force):
    dfu.build(input_file, output_file, address, force)


@dfu_cli.command(help='Dump the .bin parts of a DFU file')
@click.option('--input-file', type=pathlib.Path, help='Input .bin', required=True)
@click.option('--force/--no-force', help='Force file overwrite', default=True)
def dump(input_file, force):
    dfu.dump(input_file, force)


@dfu_cli.command(help='Read information from a DFU file')
@click.option('--input-file', type=pathlib.Path, help='Input .bin', required=True)
def read(input_file):
    parsed = dfu.read(input_file)
    dfu.display_dfu_info(parsed)



import functools
import importlib
import pathlib
import pkgutil
import re

import click

from . import builders
from .formatter import AssetFormatter
from .writer import AssetWriter


def make_symbol_name(base=None, working_path=None, input_file=None, input_type=None, input_subtype=None, prefix=None):
    if base is None:
        if input_file is None:
            raise NameError("No base name or input file provided.")
        if working_path is None:
            name = '_'.join(input_file.parts)
        else:
            name = '_'.join(input_file.relative_to(working_path).parts)
    else:
        name = base.format(
            filename=input_file.with_suffix('').name,
            filepath=input_file.with_suffix(''),
            fullname=input_file.name,
            fullpath=input_file,
            type=input_type,
            subtype=input_subtype
        )
    name = name.replace('.', '_')
    name = re.sub('[^0-9A-Za-z_]', '_', name)
    name = name.lower()

    if type(prefix) is str:
        name = prefix + name

    return name


class AssetBuilder:

    _by_name = {}
    _by_extension = {}

    def __init__(self, typemap):
        self.typemap = typemap

    def __call__(self, build_func):
        self.name = build_func.__name__
        self.build = build_func
        self._by_name[self.name] = self
        for subtype, extensions in self.typemap.items():
            for ext, auto in extensions.items():
                if auto:
                    if ext in self._by_extension:
                        raise KeyError(f'An automatic handler for {ext} has already been registered ({self._by_extension[ext]}).')
                    self._by_extension[ext] = f'{self.name}/{subtype}'
        return self

    def __repr__(self):
        return self.name

    @staticmethod
    def build(self, data, subtype, **kwargs):
        raise NotImplementedError

    def from_file(self, path, subtype, **kwargs):
        if subtype is None:
            subtype = self.guess_subtype(path)
        elif subtype not in self.typemap.keys():
            raise ValueError(f'Invalid subtype {subtype}, choices {self.typemap.keys()}')
        return self.build(path.read_bytes(), subtype, **kwargs)

    def guess_subtype(self, path):
        for input_type, extensions in self.typemap.items():
            if path.suffix in extensions:
                return input_type
        raise TypeError(f"Unable to identify type of input file {path.name}.")

    @classmethod
    def guess_builder(cls, path):
        try:
            return cls._by_extension[path.suffix]
        except KeyError:
            raise TypeError('Could not find a builder for {path}.')


class AssetTool:

    _commands = {}

    def __init__(self, builder, help):
        self.builder = builder
        self.name = builder.name
        self.help = help

    def __call__(self, f):
        @click.command(self.name, help=self.help)
        @click.option('--input_file', type=pathlib.Path, required=True, help='Input file')
        @click.option('--input_type', type=click.Choice(self.builder.typemap.keys(), case_sensitive=False), default=None, help='Input file type')
        @click.option('--output_file', type=pathlib.Path, default=None, help='Output file')
        @click.option('--output_format', type=click.Choice(AssetFormatter.names(), case_sensitive=False), default=None, help='Output file format')
        @click.option('--symbol_name', type=str, default=None, help='Output symbol name')
        @click.option('--force/--keep', default=False, help='Force file overwriting')
        @functools.wraps(f)
        def cmd(input_file, input_type, output_file, output_format, symbol_name, force, **kwargs):
            aw = AssetWriter()
            aw.add_asset(symbol_name, f(input_file, input_type, **kwargs))
            aw.write(output_format, output_file, force, report=False)

        self._commands[self.name] = cmd


# Load all the implementations dynamically.
for loader, module_name, is_pkg in pkgutil.walk_packages(builders.__path__, builders.__name__ + '.'):
    # We don't need to import anything from the modules. We just need to load them.
    # This will cause the decorators to run, which registers the builders.
    importlib.import_module(module_name, builders.__name__)
